package javato.activetesting;

import java.util.LinkedHashSet;
import java.util.HashMap;
import java.io.* ;

import javato.activetesting.common.MersenneTwisterFast;
import javato.activetesting.common.Parameters ;
import javato.activetesting.activechecker.ActiveChecker;
import javato.activetesting.activechecker.RecursiveLock;
import javato.activetesting.coverage.* ;
import javato.activetesting.analysis.CheckerAnalysisImpl;
import javato.activetesting.racefuzzer.RaceChecker;
import javato.activetesting.reentrant.IgnoreRentrantLock;

/**
 * Copyright (c) 2007-2008,
 * Koushik Sen    <ksen@cs.berkeley.edu>
 * All rights reserved.
 * <p/>
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * <p/>
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * <p/>
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 * <p/>
 * 3. The names of the contributors may not be used to endorse or promote
 * products derived from this software without specific prior written
 * permission.
 * <p/>
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


public class RandomFuzzerAnalysis extends CheckerAnalysisImpl {

	public static LinkedHashSet<SyncPair> coveredSyncPair = null ;
	public static LinkedHashSet<StmtPair> coveredStmtPair = null ;

	public static HashMap<Integer, Integer> last = new HashMap<Integer, Integer>();
	public static HashMap<Long, Integer> lastwrite = new HashMap<Long, Integer>(); 
	public static MersenneTwisterFast rand = new MersenneTwisterFast(System.currentTimeMillis());
	public static double probability = Parameters.random_probability;
	public static int maxwait = Parameters.max_wait > 0 ? Parameters.max_wait+1 : 1;
	public static int minwait = maxwait > Parameters.min_wait && Parameters.min_wait >= 0 ? Parameters.min_wait : maxwait - 1;  

	public static LinkedHashSet<Integer> interestingIids = new LinkedHashSet<Integer>() ;
	public static HashMap<Integer, RecursiveLock> wrapperLock = new HashMap<Integer, RecursiveLock>() ;
	public static RecursiveLock lockMemoryAccess = new RecursiveLock(1) ;

	public static boolean eraser = Parameters.random_eraser;
	public static boolean waitlock = Parameters.random_wait;


	public void initialize() {

		System.out.println("(debug) pb: "+probability);
		System.out.println("(debug) max: "+maxwait);
		System.out.println("(debug) min: "+minwait);
		//		
		//		for (int i = 0; i<15; i++){
		//			int time = rand.nextInt(maxwait-minwait)+minwait;
		//			System.out.println("(debug) time: "+time);
		//		}
		System.out.println("(debug) wait: "+waitlock);
		System.out.println("(debug) eraser: "+eraser);

		synchronized (ActiveChecker.lock) {
			if (Parameters.errorId >= 0) {
				/* a code clone of HybridRaceTracker.getFacesFromFile() */
				ObjectInputStream in = null ;
				try {
					in = new ObjectInputStream(new BufferedInputStream(new FileInputStream(Parameters.ERROR_LOG_FILE))) ;
					coveredSyncPair = (LinkedHashSet<SyncPair>) in.readObject() ;
					if (coveredSyncPair == null) {
						coveredSyncPair = new LinkedHashSet<SyncPair>() ;

					}
					coveredStmtPair = (LinkedHashSet<StmtPair>) in.readObject() ;
					if (coveredStmtPair == null) {
						coveredStmtPair = new LinkedHashSet<StmtPair>() ;
					}
					in.close() ;
					if(eraser){
						in = new ObjectInputStream(new BufferedInputStream(new FileInputStream(Parameters.ERASER))) ;
						interestingIids = (LinkedHashSet<Integer>) in.readObject() ;
						//System.out.println(interestingIids);
						in.close() ;
					}

				}
				catch (IOException e) {
					coveredSyncPair = new LinkedHashSet<SyncPair>() ;
					coveredStmtPair = new LinkedHashSet<StmtPair>() ;
				}
				catch (ClassNotFoundException e) {
					coveredSyncPair = new LinkedHashSet<SyncPair>() ;
					coveredStmtPair = new LinkedHashSet<StmtPair>() ;
				}
				/* load coverage metric done by previous runs */
			}
			System.out.println("(debug) SyncPair: " + coveredSyncPair.size() + " / StmtPair: " + coveredStmtPair.size()) ;
		}
	}

	public void lockBefore(Integer iid, Integer thread, Integer lock) {
		synchronized(ActiveChecker.lock){
			if(rand.nextDouble()<=probability){
				int time = rand.nextInt(maxwait-minwait)+minwait;
				if(time==0)
					Thread.yield();
				else
					(new ActiveChecker()).check(time);
			}	
		}ActiveChecker.blockIfRequired();
		RecursiveLock sem = null ;
		synchronized (wrapperLock) {
			sem = wrapperLock.get(lock) ;
			if (sem == null) {
				wrapperLock.put(lock, new RecursiveLock(1)) ;
				sem = wrapperLock.get(lock) ;
			}			
		}
		try{
			boolean rec = sem.isRecursive(thread);
			sem.acquire(thread);
			synchronized (ActiveChecker.lock) {
				Integer lastiid = null ;
				lastiid = (Integer) last.get(lock) ;
				if (lastiid == null) {
					lastiid = new Integer(-1) ;
				}
				coveredSyncPair.add(new SyncPair(lastiid, iid)) ;
				coveredStmtPair.add(new StmtPair(lastiid, iid)) ;
				if (!rec)
					last.put(lock, iid) ;
			}
		}catch(InterruptedException ie) {

		}
	}

	public void lockAfter(Integer iid, Integer thread, Integer lock) {
		/* The following code should be executed right after 
		 * the current thread holds the lock.
		/* I assume that the current thread is going to 
		 * hold the lock immediately */
	}

	public void unlockAfter(Integer iid, Integer thread, Integer lock) {
		RecursiveLock sem = wrapperLock.get(lock) ;
		sem.release(thread) ;	
	}

	public void newExprAfter(Integer iid, Integer object,
			Integer objOnWhichMethodIsInvoked) {

	}

	public void methodEnterBefore(Integer iid) {

	}

	public void methodExitAfter(Integer iid) {

	}

	public void startBefore(Integer iid, Integer parent, Integer child) {

	}

	public void waitAfter(Integer iid, Integer thread, Integer lock) {

	}

	public void notifyBefore(Integer iid, Integer thread, Integer lock) {

	}

	public void notifyAllBefore(Integer iid, Integer thread, Integer lock) {

	}

	public void joinAfter(Integer iid, Integer parent, Integer child) {

	}

	public void readBefore(Integer iid, Integer thread, Long memory) {
		synchronized (ActiveChecker.lock) {
			if (waitlock){
				boolean a = eraser ? interestingIids.contains(iid) : true;
				if(a && rand.nextDouble()<=probability){
					int time = rand.nextInt(maxwait-minwait)+minwait;
					if(time==0)
						Thread.yield();
					else
						(new ActiveChecker()).check(time);
				}
			}
		}ActiveChecker.blockIfRequired();
		try{
			lockMemoryAccess.acquire(thread);

			synchronized (ActiveChecker.lock) {
				Integer lastiid = null ;
				lastiid = (Integer) lastwrite.get(memory) ;
				if (lastiid == null) {
					lastiid = new Integer(-1) ;
				}
				coveredStmtPair.add(new StmtPair(lastiid, iid)) ;
			}
		}
		catch(InterruptedException ie) {
			System.out.println(ie);
		}
	}

	public void readAfter(Integer iid, Integer thread, Long memory) {
		lockMemoryAccess.release(thread);
	}	

	public void writeBefore(Integer iid, Integer thread, Long memory) {
		synchronized (ActiveChecker.lock) {
			if (waitlock){
				boolean a = eraser ? interestingIids.contains(iid) : true;
				if(a && rand.nextDouble()<=probability){
					int time = rand.nextInt(maxwait-minwait)+minwait;
					if(time==0)
						Thread.yield();
					else
						(new ActiveChecker()).check(time);
				}
			}
		}ActiveChecker.blockIfRequired();

		try{
			lockMemoryAccess.acquire(thread);

			synchronized (ActiveChecker.lock) {

				Integer lastiid = null ;
				lastiid = (Integer) lastwrite.get(memory) ;
				if (lastiid == null) {
					lastiid = new Integer(-1) ;
				}
				coveredStmtPair.add(new StmtPair(lastiid, iid)) ;
				lastwrite.put(memory, iid) ;
			}
		}catch(InterruptedException ie){
			System.out.println(ie) ;
			System.exit(0) ;
		}
	}

	public void writeAfter(Integer iid, Integer thread, Long memory) {
		lockMemoryAccess.release(thread);
	}	

	public void finish() {
		/* a code clone from HybridRaceTracker::dumpRaces() */
		ObjectOutputStream out = null ;
		try {
			if (coveredSyncPair == null){

				coveredSyncPair = new LinkedHashSet<SyncPair>() ;
			}
			if (coveredStmtPair == null){

				coveredStmtPair = new LinkedHashSet<StmtPair>() ;
			}

			out = new ObjectOutputStream(new BufferedOutputStream(new FileOutputStream(Parameters.ERROR_LOG_FILE))) ;
			out.writeObject(coveredSyncPair) ;
			out.writeObject(coveredStmtPair) ;
			out.close() ;
			ObjectInputStream in = new ObjectInputStream(new BufferedInputStream(new FileInputStream(Parameters.ERROR_LOG_FILE)));


			System.out.println("Number of covered SyncPair items " + coveredSyncPair.size()) ;
			System.out.println("Number of covered StmtPair items " + coveredStmtPair.size()) ;


			//			System.out.println("Number of covered StmtPair items " + coveredStmtPair.size()) ;
			//			System.out.println(StmtPairCovFuzzerAnalysis.covered) ;
			out.close() ;
		}
		catch (IOException e) {
			e.printStackTrace() ;
		}


		writeStat(Parameters.ERROR_STAT_FILE);
	}

	public static void writeStat(String file) {
		try {
			PrintWriter pw = new PrintWriter(new FileWriter(file,true));
			pw.print(Parameters.errorId+":"+RaceChecker.isRace);
			pw.close();
		} catch (IOException e) {
			System.err.println("Error while writing to " + file);
			System.exit(1);
		}
	}
}
