package javato.activetesting;

import javato.activetesting.activechecker.ActiveChecker;
import javato.activetesting.activechecker.RecursiveLock;
import javato.activetesting.activechecker.Semaphore ;
import javato.activetesting.analysis.CheckerAnalysisImpl;
import javato.activetesting.common.MersenneTwisterFast;
import javato.activetesting.common.Parameters;
import javato.activetesting.hybridracedetection.CommutativePair;
import javato.activetesting.hybridracedetection.HybridRaceTracker;
import javato.activetesting.racefuzzer.RaceChecker;
import javato.activetesting.coverage.* ;

import java.util.LinkedHashSet;
import java.util.LinkedHashMap ;
import java.util.HashMap ;
import java.util.Iterator ;
import java.util.LinkedList ;

import java.io.ObjectInputStream ;
import java.io.ObjectOutputStream ;
import java.io.BufferedInputStream ;
import java.io.BufferedOutputStream ;
import java.io.FileInputStream ;
import java.io.FileOutputStream ;
import java.io.PrintWriter;
import java.io.FileWriter;
import java.io.IOException;

/**
 * Copyright (c) 2007-2008,
 * Koushik Sen    <ksen@cs.berkeley.edu>
 * All rights reserved.
 * <p/>
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * <p/>
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * <p/>
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 * <p/>
 * 3. The names of the contributors may not be used to endorse or promote
 * products derived from this software without specific prior written
 * permission.
 * <p/>
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

public class SPTester extends CheckerAnalysisImpl {
	public static double pb_control = Parameters.feedback_probability ;
	public static MersenneTwisterFast rand = new MersenneTwisterFast(System.currentTimeMillis());
	
	public static LinkedHashSet<SyncPair> covered = new LinkedHashSet<SyncPair>() ;
	public static HashMap<Integer, Integer> last = new HashMap<Integer, Integer>() ; /* last: lock identifier -> an iid that holds the lock right before */
	public static HashMap<Integer, RecursiveLock> wrapperLock = new HashMap<Integer, RecursiveLock>() ; /* wrapperLock:  lock-id -> semaphore */

	public static LinkedHashSet<StmtPair> coveredStmtPair = new LinkedHashSet<StmtPair>() ;
	public static HashMap<Long, Integer> lastWrite = new HashMap<Long, Integer>() ;
	public static RecursiveLock wrapperLockMemory = new RecursiveLock(1) ;

	public static LinkedHashSet<SyncPair> notyetcovered = new LinkedHashSet<SyncPair>() ; 
	
	public static int countRule3_normal = 0;
	public static int countRule3_max = 0;
	
	public void initialize() {
		synchronized (ActiveChecker.lock) {
			if (Parameters.errorId >= 0) {
				/* a code clone of HybridRaceTracker.getFacesFromFile() */
				ObjectInputStream in = null ;
				try {
					pb_control = Parameters.feedback_probability ;
//					pb_control = 0.9;
					System.out.println("conf) feedback_probability: " + pb_control) ;

					in = new ObjectInputStream(new BufferedInputStream(new FileInputStream(Parameters.ERROR_LOG_FILE))) ;
					covered = (LinkedHashSet<SyncPair>) in.readObject() ;
					notyetcovered = (LinkedHashSet<SyncPair>) in.readObject() ;					
					coveredStmtPair = (LinkedHashSet<StmtPair>) in.readObject() ;

					in.close() ;

					System.out.println("# of covered SyncPair: " + covered.size()) ;
					System.out.println("# of covered StmtPair: " + coveredStmtPair.size()) ;
				}
				catch (IOException e) {
					covered = new LinkedHashSet<SyncPair>() ;
				}
				catch (ClassNotFoundException e) {
					covered = new LinkedHashSet<SyncPair>() ; 
				}
				/* load coverage metric done by previous runs */
			}
			//			System.out.println("(debug) size: " + covered.size()) ;
		}
		System.out.println("(init) ||notyetcovered||:" + notyetcovered.size()) ;
	}

	public void finish() {
		/* a code clone from HybridRaceTracker::dumpRaces() */
		ObjectOutputStream out = null ;
		try {
			synchronized (ActiveChecker.lock) {
//				Thread[] tList = new Thread[Thread.activeCount()];
//				int numThreads = Thread.enumerate(tList);
//				for (int i = 0; i<numThreads; i++){
//					System.out.println("---thread Name: "+tList[i].getName()+", ID: "+tList[i].getId());
//				}
				out = new ObjectOutputStream(new BufferedOutputStream(new FileOutputStream(Parameters.ERROR_LOG_FILE))) ;
				out.writeObject(covered) ;
				out.writeObject(notyetcovered) ;
				out.writeObject(coveredStmtPair) ;
				out.close() ;
				/* screen */
				System.out.println("Number of covered SyncPair items " + covered.size()) ;
				System.out.println("Number of covered StmtPair items " + coveredStmtPair.size()) ;
				System.out.println("Number of applyed Rule " + countRule3_normal);
				System.out.println("Number of applyed Rule and max_score " + countRule3_max);
				/*
			System.out.println("Uncovered SyncPair items:") ;
			for (Iterator<SyncPair> itr = notyetcovered.iterator() ; itr.hasNext() == true ; ) {
				System.out.print(itr.next()) ;
				System.out.print(" ") ;
			}
			System.out.println("\n====") ;
				 */
			}

		}
		catch (IOException e) {
			e.printStackTrace() ;
		}
		/* upto here */

		writeStat(Parameters.ERROR_STAT_FILE);
	}

	public static void writeStat(String file) {
		try {
			PrintWriter pw = new PrintWriter(new FileWriter(file,true));
			pw.print(Parameters.errorId+":"+RaceChecker.isRace);
			pw.close();
		} catch (IOException e) {
			System.err.println("Error while writing to " + file);
			System.exit(1);
		}
	}

	public boolean checkNotYetCovered(Integer iid) {
		Iterator<SyncPair> itr = notyetcovered.iterator() ;
		while (itr.hasNext() == true) {
			SyncPair sp = itr.next() ;
			if (sp.x.intValue() == iid.intValue() || sp.y.intValue() == iid.intValue())
				return true ;
		}
		return false ;
	}

	public boolean isUnderControl(Integer lock) {
		Iterator<ActiveChecker> itr = ActiveChecker.blockedThreads.iterator() ;
		while (itr.hasNext() == true) {
			CovActiveChecker cac = (CovActiveChecker)itr.next() ;
			if (cac.lock.intValue() == lock.intValue()) {
				return true ;
			}
		}
		return false ;
	}

	/* handlers for each type of injection */

	public void lockBefore(Integer iid, Integer thread, Integer lock) {
		//System.out.println("(d) lockBefore Thread-"+thread+ ", lock-"+lock) ;		

		RecursiveLock sem = null ;
		synchronized (ActiveChecker.lock) {
			Integer lastiid = null ;
			lastiid = (Integer) last.get(lock) ;		
			if (lastiid == null) {
				lastiid = new Integer(-1) ;
			}

			synchronized (wrapperLock) {
				sem = wrapperLock.get(lock) ;
				if (sem == null) {
					wrapperLock.put(lock, new RecursiveLock(1)) ;
					sem = wrapperLock.get(lock) ;
				}			
			}

			if (sem.isRecursive(thread) == true) {
				try {
					sem.acquire(thread) ;
					synchronized (ActiveChecker.lock) {
						Integer reclastiid = null ;
						reclastiid = (Integer) last.get(lock) ;
						if (reclastiid == null) {
							reclastiid = new Integer(-1) ;
						}
						covered.add(new SyncPair(reclastiid, iid)) ;
						coveredStmtPair.add(new StmtPair(reclastiid, iid)) ; /* StmtPair */
						notyetcovered.remove(new SyncPair(reclastiid, iid)) ;
					}
				}
				catch (InterruptedException ie) {
					System.out.println(ie) ;
					System.exit(0) ;
				}
				return ;
			}

			
			if (rand.nextDouble() < pb_control) {			
				if (/*(covered.contains(new SyncPair(lastiid, iid)) == true) ||*/ (checkNotYetCovered(iid) == true)){
					/* try to block the current thread's execution 
					* if the resulted coverage task is already satisfied
					* */				
					CovActiveChecker cac = new CovActiveChecker(iid, thread, lock, notyetcovered) ;
					cac.check() ; /* in the case for blocking */
				}
				else {
					
					if (isUnderControl(lock)) {
						CovActiveChecker cac = new CovActiveChecker(iid, thread, lock, notyetcovered) ;
						cac.check() ;
					}
					
				}
			}

		}
//		System.out.println("log__) @" + iid + " Thread-"+thread+" may wait at here for lock-" + lock) ;
		ActiveChecker.blockIfRequired() ;
//		System.out.println("log__) @" + iid + " Thread-"+thread+" now acquires lock-"+lock) ;


		synchronized (wrapperLock) {
			sem = wrapperLock.get(lock) ;
			if (sem == null) {
				wrapperLock.put(lock, new RecursiveLock(1)) ;
				sem = wrapperLock.get(lock) ;
			}			
		}
		try {			
			sem.acquire(thread) ;			
//			System.out.println("log__) @" + iid + " Thread-"+thread+" just held lock-"+lock) ;
			synchronized (ActiveChecker.lock) {
				Integer lastiid = null ;
				lastiid = (Integer) last.get(lock) ;
				if (lastiid == null) {
					lastiid = new Integer(-1) ;
				}
				covered.add(new SyncPair(lastiid, iid)) ;
				coveredStmtPair.add(new StmtPair(lastiid, iid)) ;
				notyetcovered.remove(new SyncPair(lastiid, iid)) ;
				last.put(lock, iid) ;									
			}
		}		
		catch (InterruptedException ie) {
			System.out.println(ie) ;
			System.exit(1) ;
			//			sem.releaseAll(thread) ;
		}		
	}

	public void lockAfter(Integer iid, Integer thread, Integer lock) {

	}

	public void unlockAfter(Integer iid, Integer thread, Integer lock) {
		//System.out.println("(d) " + iid + " unlockAfter Thread-"+thread+ ", lock-"+lock) ;
		RecursiveLock sem ;
		synchronized (wrapperLock) {
			sem = wrapperLock.get(lock) ;
			if (sem == null) {
				wrapperLock.put(lock, new RecursiveLock(1)) ;
				sem = wrapperLock.get(lock) ;
			}			
		}

//		if (sem.wasRecursive(thread) == true) {
//			sem.release(thread) ;
//			return ;
//		}
//
//		if (rand.nextDouble() < pb_control) {
//			synchronized (ActiveChecker.lock) {
//				Iterator<ActiveChecker> itr = ActiveChecker.blockedThreads.iterator() ;	
//				LinkedList<CovActiveChecker> candidates = new LinkedList<CovActiveChecker>() ;
//				while (itr.hasNext() == true) {
//					CovActiveChecker c = (CovActiveChecker)itr.next() ;
//					if (c.lock.intValue() == lock.intValue()) {
//						int lastiid = last.get(lock).intValue() ;			
//						if ((notyetcovered.contains(new SyncPair(lastiid, c.iid.intValue())) == true)) {
//							//System.out.println("(debug) new cover checker: " + c.thread) ;					
//							candidates.add(c) ;
//							//sem.releaseSpecificThread(thread, c.thread) ;					
//							//c.unblockControl() ;
//							//						System.out.println("log__) @" + iid + " Thread-"+thread+" now releases lock-"+lock+"w/ indicating Thread-"+c.thread) ;
//							//return ;
//						}
//					}
//				}
//				if (candidates.size() > 0) {
//					int choice = rand.nextInt(candidates.size()) ;
//					CovActiveChecker c = candidates.get(choice) ;
//					sem.releaseSpecificThread(thread, c.thread) ;					
//					c.unblockControl() ;
//					//System.out.println("(log) @" + iid + " Thread-"+thread+" now releases lock-"+lock+"w/ indicating Thread-"+c.thread) ;
//					return ;				
//				}
//				// the one that can make a least number of pairs		
//
//				/****/
//				LinkedList<CovActiveCheckerPair> candidatePairs = new LinkedList<CovActiveCheckerPair>() ;
//
//				for (int i = 0 ; i < ActiveChecker.blockedThreads.size() ; i++) {
//					for (int j = i + 1 ; j < ActiveChecker.blockedThreads.size() ; j++) {
//						CovActiveChecker a = (CovActiveChecker)ActiveChecker.blockedThreads.get(i) ;
//						CovActiveChecker b = (CovActiveChecker)ActiveChecker.blockedThreads.get(j) ;
//						if (a.lock.intValue() == lock.intValue()) {
//							if (a.lock.intValue() == b.lock.intValue()) {
//								if (notyetcovered.contains(new SyncPair(a.iid.intValue(), b.iid.intValue()))) {
//									candidatePairs.add(new CovActiveCheckerPair(a, b)) ;
//								}
//								else if (notyetcovered.contains(new SyncPair(b.iid.intValue(), a.iid.intValue()))) {
//									candidatePairs.add(new CovActiveCheckerPair(b, a)) ;
//								}
//							}
//						}
//					}
//				}
//				if (candidatePairs.size() > 0) {
//					int choice = rand.nextInt(candidatePairs.size()) ;
//					CovActiveCheckerPair cacp = candidatePairs.get(choice) ;
//					sem.releaseSpecificThread(thread, cacp.a.thread) ;				
//					cacp.a.unblockControl() ;
//					//System.out.println("(log) unblockAThread decide <" + cacp.a.iid + "," + cacp.b.iid + "> among " + candidatePairs.size()) ;
//					return ;
//				}
//				/****/		
//
//				candidates = null ;
//				itr = ActiveChecker.blockedThreads.iterator() ;
//				int minValue = -1 ;		
//				CovActiveChecker minChecker = null ;
//				while (itr.hasNext() == true) {		
//					CovActiveChecker c = (CovActiveChecker)itr.next() ;
//					if (c.lock.intValue() == lock.intValue()) { /* ? */
//						if (c.score() > 0) {
//							if ((minChecker == null) || (c.score() < minValue)) {
//								minValue = c.score() ;
//								minChecker = c ;
//								candidates = new LinkedList<CovActiveChecker>() ;
//								candidates.add(c) ;
//							}
//							else if (c.score() < minValue) {
//								candidates.add(c) ;
//							}
//						}
//					}
//				}
//				if (minChecker != null) {
//					//System.out.println("(debug) minChecekr: " + lock + "/" + maxChecker.thread) ;	
//					int choice = rand.nextInt(candidates.size()) ;
//					CovActiveChecker c = candidates.get(choice) ;
//					sem.releaseSpecificThread(thread, c.thread) ;			
//					c.unblockControl() ;	
//					//System.out.println("(log) @" + iid + " Thread-"+thread+" now releases lock-"+lock+"w/ indicating Thread-"+minChecker.thread) ;
//					return ;
//				}
//				//			System.out.println("log__) @" + iid + " Thread-"+thread+" now releases lock-"+lock) ;
//			}
//		}
		sem.release(thread) ;		
	}

	public void newExprAfter(Integer iid, Integer object, Integer objOnWhichMethodIsInvoked) {
	}

	public void methodEnterBefore(Integer iid) {
	}

	public void methodExitAfter(Integer iid) {
	}

	public void startBefore(Integer iid, Integer parent, Integer child) {
	}

	public void waitAfter(Integer iid, Integer thread, Integer lock) {
	}

	public void notifyBefore(Integer iid, Integer thread, Integer lock) {
	}

	public void notifyAllBefore(Integer iid, Integer thread, Integer lock) {
	}

	public void joinAfter(Integer iid, Integer parent, Integer child) {
	}

	public void readBefore(Integer iid, Integer thread, Long memory) {
		try {
			wrapperLockMemory.acquire(thread) ;
			synchronized (ActiveChecker.lock) {
				Integer lastiid = lastWrite.get(memory) != null ? (Integer) lastWrite.get(memory) : new Integer(-1) ;
				coveredStmtPair.add(new StmtPair(lastiid, iid)) ;
			}			
		}
		catch (InterruptedException ie) {
			System.out.println(ie) ;
			System.exit(1) ;
		}

	}

	public void readAfter(Integer iid, Integer thread, Long memory) {
		wrapperLockMemory.release(thread) ;
	}	

	public void writeBefore(Integer iid, Integer thread, Long memory) {
		try {
			wrapperLockMemory.acquire(thread) ;
			synchronized (ActiveChecker.lock) {
				Integer lastiid = lastWrite.get(memory) != null ? (Integer) lastWrite.get(memory) : new Integer(-1) ;
				coveredStmtPair.add(new StmtPair(lastiid, iid)) ;
			}		
		}
		catch (InterruptedException ie) {
			System.out.println(ie) ;
			System.exit(1) ;
		}
	}

	public void writeAfter(Integer iid, Integer thread, Long memory) {
		synchronized (ActiveChecker.lock) {
			lastWrite.put(memory, iid) ;
		}
		wrapperLockMemory.release(thread) ;
	}
}
