package javato.activetesting.activechecker;

import java.util.HashMap;
import java.util.Set ;
import java.util.Iterator ;
import javato.activetesting.common.MersenneTwisterFast;

/**
 * Copyright (c) 2007-2008,
 * Koushik Sen    <ksen@cs.berkeley.edu>
 * All rights reserved.
 * <p/>
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * <p/>
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * <p/>
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 * <p/>
 * 3. The names of the contributors may not be used to endorse or promote
 * products derived from this software without specific prior written
 * permission.
 * <p/>
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
public class RecursiveLock {
    public int counter;
    //public int waitTime;
    
    public HashMap<Integer, Integer> holders = new HashMap<Integer, Integer>();
	//public HashMap<Integer, Object> lockByThread = new HashMap<Integer, Object>() ;
	/* lockByThread: Thread -> Object */
	
	public int indicatedThread = -1 ;
	
	public static MersenneTwisterFast rand = new MersenneTwisterFast(System.currentTimeMillis());
        
    public RecursiveLock() {
        this(0);
        //waitTime = 0;
    }

    public RecursiveLock(int i) {
        if (i < 0) 
			throw new IllegalArgumentException(i + " < 0");
        counter = i;
    }

    public synchronized void release(Integer thread) {
        Integer cnt = (Integer) holders.get(thread);
       
        if (cnt.intValue() == 1){
		holders.remove(thread) ;
	       	counter++;
		indicatedThread = -1 ;
		this.notify() ;
        }
        else{
        	holders.put(thread, cnt-1);
		indicatedThread = -1 ;
        }
    }
	

    public synchronized void releaseSpecificThread(Integer thread, Integer targetThread) {
    	Integer cnt = (Integer) holders.get(thread);

    	if (cnt == 1){
    		holders.remove(thread) ;
    		counter++;	
    		indicatedThread = targetThread.intValue() ;
    		this.notifyAll() ;				
    	}
    	else{
    		holders.put(thread, cnt-1);
    		indicatedThread = -1 ;
    	}
    }

    public synchronized void acquire(Integer thread) throws InterruptedException {
        Integer cnt = (Integer) holders.get(thread);
        Thread t = Thread.currentThread();
        //System.out.println("(debug) Thread-"+thread+" try to acquire sem: "+cnt+", counter: "+counter);
        if (cnt == null || cnt == 0){			
			while (counter <= 0 || (indicatedThread != -1 && indicatedThread != thread.intValue())) {
				this.wait();			
				if (counter > 0) {
					if (indicatedThread == -1 || indicatedThread == thread.intValue()) {
						indicatedThread = -1 ;
						break ;
					}
				}
			}			
	        counter--;
		indicatedThread = -1 ;
	        holders.put(thread, 1);
        }else {
		indicatedThread = -1 ;
        	holders.put(thread, cnt+1);
        }
	//System.out.println("(debug) Thread-"+thread + "now holds the semaphore") ;
    }

    public synchronized boolean isRecursive(Integer thread) {
    	Integer cnt = (Integer) holders.get(thread);
    	Thread t = Thread.currentThread();
    	if (cnt == null || cnt == 0){
    		return false ;
    	}
    	return true ;	
    }

    public synchronized boolean wasRecursive(Integer thread) {
    	Integer cnt = (Integer) holders.get(thread);       
    	if (cnt.intValue() == 1){
    		return false ;
    	}
    	return true ;
    }

    public synchronized boolean isBusy() {
	    if (counter == 1)
		    return false ;
	    return true ;
    }

    public synchronized void decrement(int i) {
        counter = counter - i;
    }
}

